'''
Base Linked List Module.

This module implements a Linked List data structure.
'''

from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import Type, TypeVar, Generic, Any

T = TypeVar('T')

@dataclass
class LinkedList(ABC, Generic[T]):
    '''
    LinkedList Class.

    Attributes:
        __head (Type[T]): The head of the list. T is a node class.
        __tail (Type[T]): The tail of the list. T is a node class.
        __length (int): The length of the list.
        __circular (bool): Whether the list is circular.

    Methods:
        `__str__()`: Return the string representation of the list.
        `__len__()`: Return the length of the list.
        `__iter__()`: Return an iterator for the list.
        `iscircular()`: Check if the list is circular.
        `_link(node1: Type[T], node2: Type[T])`: Link two nodes.
        `_preprocess_index(index: int)`: Preprocess the index.
        `insert(data: Any, index: int)`: Insert data at the specified index.
        `append(data: Any)`: Append data to the end of the list.
        `appendleft(data: Any)`: Append data to the beginning of the list.
        `remove(index: int)`: Remove data at the specified index.
        `pop()`: Remove data from the end of the list.
        `popleft()`: Remove data from the beginning of the list.
        `get(index: int)`: Get data at the specified index.
        `index(data: Any)`: Get the index of the specified data.
        `clear()`: Clear the list.
    '''

    __head: Type[T] = None
    __tail: Type[T] = None
    __length: int = 0
    __circular: bool = False

    def __init__(self, cls: Type[T], circular: bool = False):
        '''
        Initialize the LinkedList.

        Args:
            circular (bool): Whether the list is circular.
        '''

        self.__cls = cls
        self.__head = None
        self.__tail = None
        self.__length = 0
        self.__circular = circular

    def __str__(self):
        return f"{self.__class__.__name__}({self.__head})"

    def __len__(self):
        return self.__length

    def __iter__(self):
        '''
        Return an iterator for the list.

        Returns:
            out (Iterator): The iterator for the list.

        Yields:
            out (Any): The data in the list.
        '''

        current_node = self.__head
        while current_node is not None:
            yield current_node.data
            current_node = current_node.next

    @abstractmethod
    def _link(self, node1: T, node2: T):
        '''
        Link two nodes.

        Args:
            node1 (Type[T]): The first node.
            node2 (Type[T]): The second node.
        '''


    def iscircular(self):
        '''
        Check if the list is circular.

        Returns:
            out (bool): Whether the list is circular.
        '''

        return self.__circular

    def _preprocess_index(self, index: int, insert: bool = False):
        '''
        Preprocess the index.

        Args:
            index (int): The index to preprocess.
            insert (bool): Whether the index is for insertion.
        Returns:
            out (int): The preprocessed index.
        
        Raises:
            IndexError: If the index is out of bounds.
        '''

        ins = 1 if insert else 0

        if index < 0:
            index = self.__length + index + ins

        if index < 0 or ((1 - ins + index) > self.__length):
            raise IndexError("Index out of bounds")

        return index

    def insert(self, data: Any, index: int):
        '''
        Insert data at the specified index.

        Args:
            data (Any): The data to be inserted.
            index (int): The index to insert the data.
        
        Raises:
            IndexError: If the index is out of bounds.
        '''

        index = self._preprocess_index(index, insert=True)

        if index == 0:
            self.appendleft(data)
            return

        if index == self.__length:
            self.append(data)
            return

        new_node = self.__cls(data)

        current_node = self.__head
        for _ in range(index - 1):
            current_node = current_node.next

        self._link(new_node, current_node.next)
        self._link(current_node, new_node)

        self.__length += 1

    def append(self, data: Any):
        '''
        Append data to the end of the list.

        Args:
            data (Any): The data to be appended.
        '''

        new_node = self.__cls(data)

        if self.__length == 0:
            self.__head = new_node
            self.__tail = new_node

        else:
            self._link(self.__tail, new_node)
            self.__tail = new_node

        if self.__circular:
            self._link(self.__tail, self.__head)

        self.__length += 1

    def appendleft(self, data: Any):
        '''
        Append data to the beginning of the list.

        Args:
            data (Any): The data to be appended.
        '''

        new_node = self.__cls(data)

        if self.__length == 0:
            self.__head = new_node
            self.__tail = new_node

        else:
            self._link(new_node, self.__head)
            self.__head = new_node

        if self.__circular:
            self._link(self.__tail, self.__head)

        self.__length += 1

    def remove(self, index: int):
        '''
        Remove data at the specified index.

        Args:
            index (int): The index to remove the data.

        Returns:
            out (Any): The removed data.

        Raises:
            IndexError: If the index is out of bounds.
        '''

        index = self._preprocess_index(index)

        if index == 0:
            return self.popleft()

        if index == self.__length - 1:
            return self.pop()

        current_node = self.__head
        for _ in range(index - 1):
            current_node = current_node.next

        removed_node = current_node.next
        self._link(current_node, removed_node.next)
        removed_node.next = None

        self.__length -= 1

        return removed_node.data

    def pop(self):
        '''
        Extract data from the end of the list.

        Returns:
            out (Any): The extracted data.

        Raises:
            IndexError: If the list is empty.
        '''

        if self.__length == 0:
            raise IndexError("Empty list")

        if self.__length == 1:
            removed_node = self.__head
            self.__head = None
            self.__tail = None

        else:
            current_node = self.__head
            while current_node.next != self.__tail:
                current_node = current_node.next

            removed_node = self.__tail
            self._link(current_node, removed_node.next)
            self.__tail = current_node

        self.__length -= 1

        return removed_node.data

    def popleft(self):
        '''
        Extract data from the beginning of the list.

        Returns:
            out (Any): The extracted data.

        Raises:
            IndexError: If the list is empty.
        '''

        if self.__length == 0:
            raise IndexError("Empty list")

        removed_node = self.__head

        if self.__length == 1:
            self.__head = None
            self.__tail = None

        else:
            self.__head = self.__head.next

            if self.__circular:
                self._link(self.__tail, self.__head)

        self.__length -= 1

        return removed_node.data

    def get(self, index: int):
        '''
        Get data at the specified index.

        Args:
            index (int): The index to get the data.

        Returns:
            out (Any): The data at the specified index.

        Raises:
            IndexError: If the index is out of bounds.
        '''

        index = self._preprocess_index(index)

        current_node = self.__head
        for _ in range(index):
            current_node = current_node.next

        return current_node.data

    def index(self, data: Any):
        '''
        Get the index of the first occurrence of the specified data.

        Args:
            data (Any): The data to search for.

        Returns:
            out (int): The index of the specified data.
        
        Raises:
            ValueError: If the data is not found.
        '''

        current_node = self.__head
        for i in range(self.__length):
            if current_node.data == data:
                return i

            current_node = current_node.next

        raise ValueError("Data not found")

    def clear(self):
        '''
        Remove all data from the list.
        '''

        self.__head = None
        self.__tail = None
        self.__length = 0



'''
Adjacency list representation of a graph.

This module provides the AdjacencyList class, which
represents an unweighted graph using an adjacency list.
'''

from warnings import warn
from pystrukts.graph import Graph

class AdjacencyList(Graph):
    '''
    Graph representation using adjacency list.

    Attributes:
        num_vertices (int): The number of vertices in the graph.
        directed (bool): True if the graph is directed, False otherwise.
        vertex_list (list): The adjacency list of the graph.

    Methods:
        copy() -> AdjacencyList: Create a copy of the graph.
        _copy_list(vertex_list: list) -> list: Copy the adjacency list.
        _raise_error_if_invalid_vertex(x: int) -> None: Raise an error if the vertex does not exist.
        has_edge(x: int, y: int) -> bool: Check if there is an edge between two vertices.
        neighbors(x: int) -> set: Get the neighbors of a vertex.
        add_edge(x: int, y: int, w: float = None) -> None: Add an edge between two vertices.
        remove_edge(x: int, y: int) -> None: Remove an edge between two vertices.
        add_vertex() -> None: Add an isolated vertex to the graph.
        remove_vertex(x: int) -> None: Remove a vertex from the graph.
    '''

    def __init__(self, num_vertices: int = None, vertex_list: list = None,
                 directed: bool = True) -> None:
        '''
        Initialize the adjacency list.

        Args:
            num_vertices (int): The number of vertices in the graph.
            vertex_list (list): The adjacency list of the graph.
            directed (bool): True if the graph is directed, False otherwise.
        '''
        if not vertex_list and not num_vertices:
            raise ValueError('You must specify either the number of vertices \
                             or the adjacency list.')

        if num_vertices and num_vertices < 0:
            raise ValueError('The number of vertices must be non-negative.')

        if vertex_list and num_vertices:
            warn('The number of vertices will be ignored, as an adjacency list was provided.')

        self.num_vertices = len(vertex_list) if vertex_list else int(num_vertices)
        self.directed = directed
        self._copy_list(vertex_list)

    def copy(self) -> 'AdjacencyList':
        '''
        Create a copy of the graph.

        Returns:
            out (AdjacencyList): The copy of the graph.
        '''

        return AdjacencyList(vertex_list = self.vertex_list, directed = self.directed)

    def _copy_list(self, vertex_list: list) -> list:
        '''
        Copy the adjacency list with floating point values.

        If the graph is undirected and the list is not
        symmetric,the function creates a symmetric list.

        Args:
            vertex_list (list): The list to be copied.
            directed (bool): True if the graph is directed, False otherwise.

        Returns:
            out (list): The list. Each element is a set of vertices.

        Raises:
            ValueError: If a vertex does not exist.
        '''

        self.vertex_list = [set() for _ in range(self.num_vertices)]

        if vertex_list:
            for u in range(self.num_vertices):
                for v in vertex_list[u]:
                    self._raise_error_if_invalid_vertex(v)
                    self.vertex_list[u].add(v)

                    if not self.directed:
                        self.vertex_list[v].add(u)

    def  _raise_error_if_invalid_vertex(self, x: int) -> None:
        '''
        Raise an error if the vertex does not exist.

        Args:
            x (int): The vertex.

        Raises:
            ValueError: If the vertex does not exist.
        '''
        if x >= self.num_vertices or x < 0 or not isinstance(x, int):
            raise ValueError(f'The vertex {x} does not exist.')

    def has_edge(self, x: int, y: int) -> bool:
        '''
        Check if there is an edge between two vertices. 
        If the vertices are the same, the function checks if the vertex has a self-loop.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.

        Returns:
            out (bool): True if there is an edge between the vertices, False otherwise.

        Raises:
            ValueError: If x or y are not valid vertices.
        '''

        self._raise_error_if_invalid_vertex(x)
        self._raise_error_if_invalid_vertex(y)

        return y in self.vertex_list[x]

    def neighbors(self, x: int) -> set:
        '''
        Get the neighbors of a vertex. 
        If x has a self-loop, the output contains the vertex itself.

        Args:
            x (int): The vertex.

        Returns:
            out (set): The set of neighbors of the vertex.

        Raises:
            ValueError: If x is not a valid vertex.
        '''

        self._raise_error_if_invalid_vertex(x)

        return self.vertex_list[x]

    def add_edge(self, x: int, y: int, w: float = None) -> None:
        '''
        Add an edge between two vertices.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.
            w (float): The weight. Not used.

        Raises:
            ValueError: If x or y are not valid vertices.
        '''
        if w:
            warn('The weight will be ignored, as the graph is unweighted.')

        self._raise_error_if_invalid_vertex(x)
        self._raise_error_if_invalid_vertex(y)

        self.vertex_list[x].add(y)

        if not self.directed:
            self.vertex_list[y].add(x)

    def remove_edge(self, x: int, y: int) -> None:
        '''
        Remove an edge between two vertices.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.

        Raises:
            ValueError: If x or y are not valid vertices.
        '''

        self._raise_error_if_invalid_vertex(x)
        self._raise_error_if_invalid_vertex(y)

        self.vertex_list[x].discard(y)

        if not self.directed:
            self.vertex_list[y].discard(x)

    def add_vertex(self) -> None:
        '''
        Add an isolated vertex to the graph.
        '''

        self.num_vertices += 1

        self.vertex_list.append(set())

    def remove_vertex(self, x: int) -> None:
        '''
        Remove a vertex from the graph. The vertices with
        higher indexes are shifted one position to the left.

        Args:
            x (int): The vertex to be removed.

        Raises:
            ValueError: If the vertex does not exist.
        '''

        self._raise_error_if_invalid_vertex(x)

        for u in range(self.num_vertices):
            self.vertex_list[u].discard(x)

        self.vertex_list.pop(x)
        self.num_vertices -= 1

        for u in range(self.num_vertices):
            self.vertex_list[u].discard(x)
            for v in range(x + 1, self.num_vertices + 1):
                if v in self.vertex_list[u]:
                    self.vertex_list[u].discard(v)
                    self.vertex_list[u].add(v - 1)



'''
Binary Search Tree.

This module contains the implementation of a binary search tree.
'''

from typing import Any
from .binary_tree import BinaryTree

class BinarySearchTree(BinaryTree):
    '''
    Binary search tree class.

    This class represents a binary search tree. There are no duplicate nodes in the tree.
    '''

    def insert(self, data: Any) -> None:
        '''
        Inserts a new node in the tree.

        Args:
            data (Any): The data of the new node.
        '''
        if data < self.data:
            if self.left is None:
                self.left = BinarySearchTree(data)
            else:
                self.left.insert(data)
        elif data > self.data:
            if self.right is None:
                self.right = BinarySearchTree(data)
            else:
                self.right.insert(data)

    def search(self, data: Any) -> bool:
        '''
        Searches for a node in the tree.

        Args:
            data (Any): The data to be searched.

        Returns:
            bool: True if the data is found, False otherwise.
        '''
        if self.data == data:
            return True

        child = self.left if data < self.data else self.right

        if child is None:
            return False

        return child.search(data)

    def _find_successor(self) -> 'BinarySearchTree':
        '''
        Finds the in-order successor of the node.

        Returns:
            BinarySearchTree: The in-order successor of the node.
        '''
        temp = self.right

        while temp.left is not None:
            temp = temp.left

        return temp

    def delete(self, data: Any) -> 'BinarySearchTree':
        '''
        Returns the root of the tree after deleting a node.

        Args:
            data (Any): The data of the node to be deleted.

        Returns:
            BinarySearchTree: The root of the tree after the deletion.
        '''
        if data < self.data:
            if self.left is not None:
                self.left = self.left.delete(data)
        elif data > self.data:
            if self.right is not None:
                self.right = self.right.delete(data)
        else:
            if self.left is None:
                return self.right
            if self.right is None:
                return self.left

            temp = self._find_successor()
            self.data = temp.data
            self.right = self.right.delete(temp.data)

        return self



'''
MaxHeap Module.

This module implements a MaxHeap data structure using a list and the heapq library.
It provides methods for inserting, extracting, and peeking at the maximum value in the heap.
'''

from dataclasses import field, dataclass
from typing import Any
import heapq

@dataclass
class MaxHeap:
    '''
    MaxHeap Class.

    Attributes:
        data (list): The internal list that stores the heap's elements.

    Methods:
        `__post_init__()`: Creates the heap from the data.
        `__str__()`: Return the string representation of the heap.
        `push(value: Any)`: Insert a value in the heap.
        `pop()`: Extract the maximum value from the heap.
        `peek()`: Peek the maximum value in the heap.
    '''

    data: list = field(default_factory=list)

    def __post_init__(self):
        '''
        Creates the heap from the data.
        '''
        self.data = [-value for value in self.data]
        heapq.heapify(self.data)

    def __str__(self):
        return f"MaxHeap({[-val for val in self.data]})"

    def push(self, value: Any):
        '''
        Insert a value in the heap.

        Args:
            value (Any): The value to be inserted.
        
        Raises:
            TypeError: If the value is not comparable.
        '''

        heapq.heappush(self.data, -value)

    def pop(self):
        '''
        Extract the maximum value from the heap.

        Returns:
            out (Any): The maximum value in the heap.

        Raises:
            IndexError: If the heap is empty.
        '''
        return -heapq.heappop(self.data)

    def peek(self):
        '''
        Peek the maximum value in the heap.

        Returns:
            out (Any): The maximum value in the heap or `None` if the heap is empty.
        '''
        return -self.data[0]



import pytest
from pystrukts.tree import BinaryTree

def test_initialization():
    tree = BinaryTree(1)
    assert tree.data == 1
    assert tree.left is None
    assert tree.right is None

    tree2 = BinaryTree(2)
    assert tree2.data == 2
    assert tree2.left is None
    assert tree2.right is None

def test_initialization_error():
    with pytest.raises(TypeError):
        BinaryTree()

def test_add_child():
    tree = BinaryTree(1)
    tree.left = BinaryTree(2)
    tree.right = BinaryTree(3)

    assert tree.left.data == 2
    assert tree.right.data == 3
    

def test_remove_child():
    tree = BinaryTree(1)
    tree.left = BinaryTree(2)
    tree.right = BinaryTree(3)

    del tree.left
    assert tree.left is None

    del tree.right
    assert tree.right is None

def test_is_leaf():
    tree = BinaryTree(1)
    assert tree.is_leaf() == True

    tree.left = BinaryTree(2)
    assert tree.is_leaf() == False

    tree.left.left = BinaryTree(3)
    assert tree.left.is_leaf() == False
    assert tree.is_leaf() == False

    del tree.left.left
    assert tree.left.is_leaf() == True
    assert tree.is_leaf() == False

    tree.left = None
    assert tree.is_leaf() == True


'''
Generic tree module.

This module contains the implementation of a generic tree with an optional number of children.
'''

from dataclasses import dataclass
from warnings import warn
from typing import Any

@dataclass
class GenericTree:
    '''
    Generic tree class.

    This class represents a generic tree with an optional number of children.
    '''
    def __init__(self, data: Any, max_children: int = 0):
        '''
        Initializes the tree. If max_children is less than 2,
        there are no restrictions on the number of children.

        Args:
            data (Any): The data of the root node.
            max_children (int): The maximum number of children the node can have.
        '''
        self.data = data
        self.children = []

        if max_children == 2:
            warn("Binary trees are better represented using the BinaryTree class.")

        self.max_children = int(max_children) if max_children > 1 else None

    def n_children(self) -> int:
        '''
        Returns the number of children of the node.

        Returns:
            int: The number of children of the node.
        '''
        return len(self.children)

    def add_child(self, data: Any) -> None:
        '''
        Adds a child to the node.

        Args:
            child (GenericTree): The child to be added.
        '''

        if self.max_children is not None and self.n_children() >= self.max_children:
            warn(f"Child not appended as node already has {self.max_children} children.")
        else:
            new_node = GenericTree(data, self.max_children)
            self.children.append(new_node)

    def remove_child(self, index: int) -> None:
        '''
        Removes a child from the node.

        Args:
            index (int): The index of the child to be removed.
        '''
        self.children.pop(index)

    def get_children(self) -> list:
        '''
        Returns the children of the node.

        Returns:
            list: The children of the node.
        '''
        return self.children.copy()

    def get_child(self, index: int) -> 'GenericTree':
        '''
        Returns a child of the node.

        Args:
            index (int): The index of the child to be returned.

        Returns:
            GenericTree: The child of the node.
        '''
        return self.children[index]

    def is_leaf(self) -> bool:
        '''
        Returns whether the node is a leaf or not.

        Returns:
            bool: True if the node is a leaf, False otherwise.
        '''
        return self.n_children() == 0



import pytest
from pystrukts.graph import AdjacencyList

def test_initilization():
    vertex_list = [[1], [2], [0]]
    graph = AdjacencyList(vertex_list=vertex_list)
    assert graph.num_vertices == 3
    assert graph.directed == True
    assert graph.vertex_list == [{1}, {2}, {0}]

    graph2 = AdjacencyList(vertex_list=vertex_list, directed=False)
    assert graph2.num_vertices == 3
    assert graph2.directed == False
    assert graph2.vertex_list == [{1, 2}, {0, 2}, {0, 1}]

    graph3 = AdjacencyList(3)
    assert graph3.num_vertices == 3
    assert graph3.directed == True
    assert graph3.vertex_list == [set(), set(), set()]

    # num_vertices ignored
    with pytest.warns(UserWarning):
        graph4 = AdjacencyList(2, vertex_list=vertex_list)
        assert graph4.num_vertices == 3
        assert graph4.directed == True
        assert graph4.vertex_list == [{1}, {2}, {0}]

def test_initialization_error():
    with pytest.raises(ValueError):
        graph = AdjacencyList(-1)

    with pytest.raises(ValueError):
        vertex_list = [[0, 1, 2], [3], [6, 7, 8]]
        graph = AdjacencyList(vertex_list=vertex_list)

    with pytest.raises(ValueError):
        graph = AdjacencyList()

def test_copy():
    vertex_list = [[1], [2], [0]]
    graph = AdjacencyList(vertex_list=vertex_list)
    graph2 = graph.copy()
    assert graph2.num_vertices == 3
    assert graph2.directed == True
    assert graph2.vertex_list == [{1}, {2}, {0}]

def test_has_edge():
    vertex_list = [[1], [], [0, 2]]

    graph = AdjacencyList(vertex_list=vertex_list)
    assert graph.has_edge(0, 0) == False
    assert graph.has_edge(0, 1) == True
    assert graph.has_edge(0, 2) == False
    assert graph.has_edge(1, 0) == False
    assert graph.has_edge(1, 1) == False
    assert graph.has_edge(1, 2) == False
    assert graph.has_edge(2, 0) == True
    assert graph.has_edge(2, 1) == False
    assert graph.has_edge(2, 2) == True

def test_has_edge_error():
    vertex_list = [[1], [2], [0]]
    graph = AdjacencyList(vertex_list=vertex_list)

    with pytest.raises(ValueError):
        graph.has_edge(0, 3)

    with pytest.raises(ValueError):
        graph.has_edge(-1, 0)

def test_neighbors():
    vertex_list = [[1], [], [0, 2]]

    graph = AdjacencyList(vertex_list=vertex_list)
    assert graph.neighbors(0) == {1}
    assert graph.neighbors(1) == set()
    assert graph.neighbors(2) == {0, 2}

def test_neighbors_error():
    vertex_list = [[1], [], [0, 2]]
    graph = AdjacencyList(vertex_list=vertex_list)

    with pytest.raises(ValueError):
        graph.neighbors(3)

    with pytest.raises(ValueError):
        graph.neighbors(-1)

    with pytest.raises(ValueError):
        graph.neighbors(1.3)

def test_add_edge():
    vertex_list = [[1], [], [0, 2]]

    graph = AdjacencyList(vertex_list=vertex_list)

    assert graph.has_edge(0, 2) == False
    graph.add_edge(0, 2)
    assert graph.has_edge(0, 2) == True
    assert graph.vertex_list == [{1, 2}, set(), {0, 2}]

    with pytest.warns(UserWarning):
        assert graph.has_edge(1, 2) == False
        graph.add_edge(1, 2, 25)
        assert graph.has_edge(1, 2) == True
        assert graph.vertex_list == [{1, 2}, {2}, {0, 2}]

def test_add_edge_undirected():
    vertex_list = [[1], [], [0, 2]]

    graph = AdjacencyList(3, directed=False)

    assert graph.has_edge(0, 2) == False
    assert graph.has_edge(2, 0) == False
    graph.add_edge(0, 2)
    assert graph.has_edge(0, 2) == True
    assert graph.has_edge(2, 0) == True


    with pytest.warns(UserWarning):
        assert graph.has_edge(1, 2) == False
        assert graph.has_edge(2, 1) == False
        graph.add_edge(1, 2, 25)
        assert graph.has_edge(1, 2) == True
        assert graph.has_edge(2, 1) == True

def test_add_edge_error():
    vertex_list = [[1], [], [0, 2]]
    graph = AdjacencyList(vertex_list=vertex_list)

    with pytest.raises(ValueError):
        graph.add_edge(0, 3)

    with pytest.raises(ValueError):
        graph.add_edge(-1, 0)

def test_remove_edge():
    vertex_list = [[1], [], [0, 2]]

    graph = AdjacencyList(vertex_list=vertex_list)

    assert graph.has_edge(0, 1) == True
    graph.remove_edge(0, 1)
    assert graph.has_edge(0, 1) == False
    assert graph.vertex_list == [set(), set(), {0, 2}]

    assert graph.has_edge(2, 0) == True
    graph.remove_edge(2, 0)
    assert graph.has_edge(2, 0) == False
    assert graph.vertex_list == [set(), set(), {2}]

def test_remove_edge_undirected():
    vertex_list = [[0, 1], [0, 1], []]
    graph = AdjacencyList(vertex_list=vertex_list, directed=False)

    assert graph.has_edge(0, 1) == True
    assert graph.has_edge(1, 0) == True
    graph.remove_edge(0, 1)
    assert graph.has_edge(0, 1) == False
    assert graph.has_edge(1, 0) == False
    assert graph.vertex_list == [{0}, {1}, set()]

    graph.remove_edge(0, 0)
    assert graph.vertex_list == [set(), {1}, set()]

def test_remove_edge_error():
    vertex_list = [[1], [], [0, 2]]
    graph = AdjacencyList(vertex_list=vertex_list)

    with pytest.raises(ValueError):
        graph.remove_edge(0, 3)

    with pytest.raises(ValueError):
        graph.remove_edge(-1, 0)

def test_add_vertex():
    vertex_list = [[1], [], [0, 2]]

    graph = AdjacencyList(vertex_list=vertex_list)
    assert graph.num_vertices == 3
    assert graph.vertex_list == [{1}, set(), {0, 2}]

    graph.add_vertex()
    assert graph.num_vertices == 4
    assert graph.vertex_list == [{1}, set(), {0, 2}, set()]

def test_remove_vertex():
    vertex_list = [[1], [], [0, 2]]

    graph = AdjacencyList(vertex_list=vertex_list)
    assert graph.num_vertices == 3
    assert graph.vertex_list == [{1}, set(), {0, 2}]

    graph.remove_vertex(1)
    assert graph.num_vertices == 2
    assert graph.vertex_list == [set(), {0, 1}]

    graph.remove_vertex(0)
    assert graph.num_vertices == 1
    assert graph.vertex_list == [{0}]

def test_remove_vertex_error():
    vertex_list = [[1], [], [0, 2]]
    graph = AdjacencyList(vertex_list=vertex_list)

    with pytest.raises(ValueError):
        graph.remove_vertex(3)

    with pytest.raises(ValueError):
        graph.remove_vertex(-1)



'''
Union-Find Disjoint Set module.

This module implements a Union-Find Disjoint Set data structure.
'''

class UnionFind:
    '''
    Union-Find Disjoint Set Class.

    Attributes:
        parents (list): The parent nodes of the elements.
        ranks (list): The ranks of the elements.
        sizes (list): The sizes of the sets.
        numdisjoint (int): The number of disjoint sets.

    Methods:
        `find(x: int)`: Find the parent of the element.
        `union(a: int, b: int)`: Union two elements.
        `size(x: int)`: Get the size of the set containing the element.
    '''

    def __init__(self, n):
        '''
        Initialize the UFDS.

        Args:
            n (int): The number of elements.
        '''

        self.parents = list(range(n))
        self.ranks = [0] * n
        self.sizes = [1] * n
        self.numdisjoint = n

    def find(self, x):
        '''
        Find the parent of the element.

        Args:
            x (int): The element.

        Returns:
            int: The parent of the element.
        '''

        x_parent = x
        children = []
        while x_parent != self.parents[x_parent]:
            children.append(x_parent)
            x_parent = self.parents[x_parent]
        for c in children:
            self.parents[c] = x_parent
        return x_parent

    def union(self, a, b):
        '''
        Union two elements.

        Args:
            a (int): The first element.
            b (int): The second element.
        '''

        a_parent = self.find(a)
        b_parent = self.find(b)

        if a_parent == b_parent:
            return

        if self.ranks[a_parent] < self.ranks[b_parent]:
            self.parents[a_parent] = b_parent
            self.sizes[b_parent] += self.sizes[a_parent]

        elif self.ranks[b_parent] < self.ranks[a_parent]:
            self.parents[b_parent] = a_parent
            self.sizes[a_parent] += self.sizes[b_parent]

        else:
            self.parents[b_parent] = a_parent
            self.ranks[a_parent] += 1
            self.sizes[a_parent] += self.sizes[b_parent]

        self.numdisjoint -= 1

    def size(self, x):
        '''
        Get the size of the set containing the element.

        Args:
            x (int): The element.

        Returns:
            int: The size of the set containing the element.
        '''

        return self.sizes[self.find(x)]



# pylint: skip-file

from .trie import Trie, TrieNode
from .generic_tree import GenericTree
from .binary_tree import BinaryTree
from .binary_search_tree import BinarySearchTree
from .fenwick_tree import FenwickTree, RUPQ, RURQ


import pytest
from pystrukts.tree import GenericTree

def test_initialization():
    tree = GenericTree(1)
    assert tree.data == 1
    assert tree.max_children is None
    assert tree.n_children() == 0

    tree2 = GenericTree(2, 3)
    assert tree2.data == 2
    assert tree2.max_children == 3
    assert tree2.n_children() == 0

    with pytest.warns(UserWarning):
        tree3 = GenericTree(3, 2)
        assert tree3.data == 3
        assert tree3.max_children == 2
        assert tree3.n_children() == 0

def test_initialization_error():
    with pytest.raises(TypeError):
        GenericTree()

def test_add_child():
    tree = GenericTree(1, 3)
    tree.add_child(2)
    assert tree.n_children() == 1
    assert tree.get_child(0).data == 2

    tree.add_child(3)
    assert tree.n_children() == 2
    assert tree.get_child(1).data == 3

    tree.add_child(4)
    assert tree.n_children() == 3
    assert tree.get_child(2).data == 4

    with pytest.warns(UserWarning):
        tree.add_child(4)
        assert tree.n_children() == 3

    tree.get_child(0).add_child(5)
    assert tree.get_child(0).n_children() == 1
    assert tree.get_child(0).get_children()[0].data == 5

def test_remove_child():
    tree = GenericTree(1, 3)
    tree.add_child(2)
    tree.add_child(3)
    tree.add_child(4)

    tree.remove_child(1)
    assert tree.n_children() == 2
    assert tree.get_child(1).data == 4

    tree.remove_child(0)
    assert tree.n_children() == 1
    assert tree.get_child(0).data == 4

    tree.remove_child(0)
    assert tree.n_children() == 0

def test_get_child():
    tree = GenericTree(1, 3)
    tree.add_child(2)
    tree.add_child(3)
    tree.add_child(4)

    assert tree.get_child(0).data == 2
    assert tree.get_child(1).data == 3
    assert tree.get_child(2).data == 4

def test_negative_index():
    tree = GenericTree(1, 3)
    tree.add_child(2)
    tree.add_child(3)
    tree.add_child(4)

    tree.remove_child(-1)
    assert tree.n_children() == 2
    assert tree.get_child(-1).data == 3

def test_remove_out_of_bounds():
    tree = GenericTree(1, 3)
    tree.add_child(2)
    tree.add_child(3)
    tree.add_child(4)

    with pytest.raises(IndexError):
        tree.remove_child(3)

    with pytest.raises(IndexError):
        tree.remove_child(-4)

def test_get_child_out_of_bounds():
    tree = GenericTree(1, 3)
    tree.add_child(2)
    tree.add_child(3)
    tree.add_child(4)

    with pytest.raises(IndexError):
        tree.get_child(3)

    with pytest.raises(IndexError):
        tree.get_child(-4)

def test_get_children():
    tree = GenericTree(1, 3)
    tree.add_child(2)
    tree.add_child(3)
    tree.add_child(4)

    children = tree.get_children()
    assert children[0].data == 2
    assert children[1].data == 3
    assert children[2].data == 4

def test_is_leaf():
    tree = GenericTree(1, 3)
    assert tree.is_leaf() == True

    tree.add_child(2)
    assert tree.is_leaf() == False

    tree.get_child(0).add_child(3)
    assert tree.get_child(0).is_leaf() == False
    assert tree.is_leaf() == False

    tree.get_child(0).remove_child(0)
    assert tree.get_child(0).is_leaf() == True
    assert tree.is_leaf() == False

    tree.remove_child(0)
    assert tree.is_leaf() == True


'''
Fenwick Tree implementation in Python.

This module contains the implementation of Fenwick Trees, as well as two extensions of it:
- Range Update Point Query (RUPQ)
- Range Update Range Query (RURQ)
'''

from typing import List

class FenwickTree:
    '''
    Fenwick Tree implementation in Python.

    Attributes:
    - n: int - the size of the Fenwick Tree
    - ftree: list - the Fenwick Tree itself

    Methods:
    - lsone(s: int) -> int: returns the least significant bit of s
    - query(i: int, j: int) -> int: returns the sum of the elements in the range [i, j]
    - update(i: int, v: int): updates the element at index i with value v
    - select(k: int) -> int: returns the index of the k-th element in the Fenwick Tree
    '''
    def __init__(self, f: List[int]):
        self.n = len(f)
        self.ftree = [0] * (self.n + 1)

        for i in range(1, self.n + 1):
            self.ftree[i] += f[i - 1]
            if i + self._lsone(i) <= self.n:
                self.ftree[i + self._lsone(i)] += self.ftree[i]

    def _lsone(self, s: int) -> int:
        '''
        Returns the least significant bit of s.

        Args:
            s (int): The number to get the least significant bit from.

        Returns:
            out (int): The least significant bit of s.
        '''
        return s & -s

    def query(self, i: int, j: int) -> int:
        '''
        Queries the Fenwick Tree for the sum of the elements in the range [i, j].

        Args:
            i (int): The lower bound of the range.
            j (int): The upper bound of the range.

        Returns:
            s (int): The sum of the elements in the range [i, j].
        '''
        if i > 1:
            return self.query(1, j) - self.query(1, i - 1)

        s = 0
        while j > 0:
            s += self.ftree[j]
            j -= self._lsone(j)

        return s

    def update(self, i: int, v: int):
        '''
        Updates the element at index i with value v.

        Args:
            i (int): The index of the element to update.
            v (int): The new value of the element.
        '''

        while i <= self.n:
            self.ftree[i] += v
            i += self._lsone(i)

    def select(self, k: int) -> int:
        '''
        Returns the index of the k-th element in the Fenwick Tree.

        Args:
            k (int): The index of the element to return.

        Returns:
            i (int): The index of the k-th element in the Fenwick Tree.
        '''

        p = 1
        while p*2 <= self.n:
            p *= 2

        i = 0
        while p > 0:
            if k > self.ftree[i + p]:
                k -= self.ftree[i + p]
                i += p
            p //= 2

        return i + 1

class RUPQ:
    '''
    Range Update Point Query (RUPQ) implementation.

    This class extends the Fenwick Tree to support range updates and point queries.

    Attributes:
    - ftree: FenwickTree - the Fenwick Tree used for the range updates

    Methods:
    - query(i: int) -> int: returns the value of the element at index i
    - update(i: int, j: int, v: int): updates the elements in the range [i, j] with value v
    '''
    def __init__(self, n: int):
        self.ftree = FenwickTree([0] * n)

    def query(self, i: int) -> int:
        '''
        Queries the Fenwick Tree for the value of the element at index i.
        
        Args:
            i (int): The index of the element to query.

        Returns:
            s (int): The value of the element at index i.
        '''

        return self.ftree.query(1, i)

    def update(self, i: int, j: int, v: int):
        '''
        Updates the elements in the range [i, j] with value v.

        Args:
            i (int): The lower bound of the range.
            j (int): The upper bound of the range.
            v (int): The value to update the elements with.
        '''

        self.ftree.update(i, v)
        self.ftree.update(j + 1, -v)

class RURQ:
    '''
    Range Update Range Query (RURQ) implementation.

    This class extends the Fenwick Tree to support range updates and range queries.

    Attributes:
    - f: FenwickTree - the Fenwick Tree used for the range updates
    - r: RUPQ - the RUPQ used for the range queries

    Methods:
    - query(i: int, j: int) -> int: returns the sum of the elements in the range [i, j]
    - update(i: int, j: int, v: int): updates the elements in the range [i, j] with value v
    '''

    def __init__(self, n: int):
        self.ftree = FenwickTree([0] * n)
        self.rupq = RUPQ(n)

    def query(self, i: int, j: int) -> int:
        '''
        Queries the Fenwick Tree for the sum of the elements in the range [i, j].

        Args:
            i (int): The lower bound of the range.
            j (int): The upper bound of the range.

        Returns:
            s (int): The sum of the elements in the range [i, j].
        '''

        if i > 1:
            return self.query(1, j) - self.query(1, i - 1)
        return self.rupq.query(j) * j - self.ftree.query(1, j)

    def update(self, i: int, j: int, v: int):
        '''
        Updates the elements in the range [i, j] with value v.

        Args:
            i (int): The lower bound of the range.
            j (int): The upper bound of the range.
            v (int): The value to update the elements with.
        '''

        self.rupq.update(i, j, v)
        self.ftree.update(i, v * (i - 1))
        self.ftree.update(j + 1, -1 * v * j)



'''
Adjacency matrix representation of a graph.

This module provides the AdjacencyMatrix class, which represents a graph using an adjacency matrix.
'''

from warnings import warn
from pystrukts.graph import Graph

class AdjacencyMatrix(Graph):
    '''
    Graph representation using adjacency matrix.

    Attributes:
        num_vertices (int): The number of vertices in the graph.
        directed (bool): True if the graph is directed, False otherwise.
        matrix (list): The adjacency matrix of the graph.
        no_edge_value (float): The value used to represent the absence of an edge.

    Methods:
        copy() -> AdjacencyMatrix: Create a copy of the graph.
        _copy_matrix(matrix: list, directed: bool) -> list: Copy the adjacency matrix.
        _raise_error_if_invalid_vertex(x: int) -> None: Raise an error if the vertex does not exist.
        has_edge(x: int, y: int) -> bool: Check if there is an edge between two vertices.
        neighbors(x: int) -> list: Get the neighbors of a vertex.
        add_edge(x: int, y: int, w: float = 1.0) -> None: Add an edge between two vertices.
            If the edge already exists, the weight is updated.
        remove_edge(x: int, y: int) -> None: Remove an edge between two vertices.
        add_vertex() -> None: Add an isolated vertex to the graph.
        remove_vertex(x: int) -> None: Remove a vertex from the graph.
        get_edge_weight(x: int, y: int) -> float: Get the weight of an edge.
        set_edge_weight(x: int, y: int, w: float) -> None: Set the weight of an edge.
    '''

    def __init__(self, num_vertices: int = None, directed: bool = True,
                 matrix: list = None, no_edge_value: float = float('inf')) -> None:
        '''
        Initialize the adjacency matrix.

        Args:
            num_vertices (int): The number of vertices in the graph.
            directed (bool): True if the graph is directed, False otherwise.
            matrix (list): The adjacency matrix of the graph.
            no_edge_value (float): The value used to represent the absence of an edge.

        Raises:
            ValueError: No adjacency matrix or number of vertices provided.
        '''

        if not matrix and not num_vertices:
            raise ValueError('You must specify either the number of vertices \
                             or the adjacency matrix.')

        if num_vertices and num_vertices < 0:
            raise ValueError('The number of vertices must be a non-negative integer.')

        if matrix and num_vertices:
            warn('The number of vertices will be ignored, as an adjacency matrix was provided.')

        self.directed = directed
        self.no_edge_value = no_edge_value

        if matrix is not None:
            for row in matrix:
                if len(row) != len(matrix):
                    raise ValueError('The number of columns in the matrix must be \
                                      equal to the number of vertices.')

            self.num_vertices = len(matrix)
            self.matrix = self._copy_matrix(matrix)
        else:
            self.num_vertices = int(num_vertices)
            self.matrix = [[no_edge_value for _ in range(num_vertices)]
                           for _ in range(num_vertices)]

    def copy(self) -> 'AdjacencyMatrix':
        '''
        Create a copy of the graph.

        Returns:
            out (AdjacencyMatrix): The copy of the graph.
        '''

        return AdjacencyMatrix(directed = self.directed,
                               matrix = self.matrix, no_edge_value = self.no_edge_value)

    def _copy_matrix(self, matrix: list) -> list:
        '''
        Copy the adjacency matrix with floating point values. 
        If the graph is undirected, the matrix is made symmetric by copying the upper triangle.


        Args:
            matrix (list): The matrix to be copied.

        Returns:
            out (list): The matrix.
        '''

        if not self.directed:
            return [[matrix[i][j] if i <= j else matrix[j][i]
                 for j in range(self.num_vertices)] for i in range(self.num_vertices)]
        return [[float(matrix[i][j]) for j in range(self.num_vertices)]
                for i in range(self.num_vertices)]

    def  _raise_error_if_invalid_vertex(self, x: int) -> None:
        '''
        Raise an error if the vertex does not exist.

        Args:
            x (int): The vertex.

        Raises:
            ValueError: If the vertex does not exist.
        '''
        if x >= self.num_vertices or x < 0 or not isinstance(x, int):
            raise ValueError(f'The vertex {x} does not exist.')

    def has_edge(self, x: int, y: int) -> bool:
        '''
        Check if there is an edge between two vertices. 
        If the vertices are the same, the function checks if the vertex has a self-loop.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.

        Returns:
            out (bool): True if there is an edge between the vertices, False otherwise.

        Raises:
            ValueError: If x or y are not valid vertices.
        '''

        self._raise_error_if_invalid_vertex(x)
        self._raise_error_if_invalid_vertex(y)

        return self.matrix[x][y] != self.no_edge_value

    def neighbors(self, x: int) -> list:
        '''
        Get the neighbors of a vertex. 
        If x has a self-loop, the output contains the vertex itself.

        Args:
            x (int): The vertex.

        Returns:
            out (list): The list of neighbors of the vertex.

        Raises:
            ValueError: If x is not a valid vertex.
        '''

        self._raise_error_if_invalid_vertex(x)

        return [i for i in range(self.num_vertices) if self.matrix[x][i] != self.no_edge_value]

    def add_edge(self, x: int, y: int, w: float = 1.0) -> None:
        '''
        Add an edge between two vertices. If the edge already exists, the weight is updated.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.
            w (float): The weight of the edge.

        Raises:
            ValueError: If x or y are not valid vertices.
        '''

        self._raise_error_if_invalid_vertex(x)
        self._raise_error_if_invalid_vertex(y)

        self.matrix[x][y] = float(w)

        if not self.directed:
            self.matrix[y][x] = float(w)

    def remove_edge(self, x: int, y: int) -> None:
        '''
        Remove an edge between two vertices.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.

        Raises:
            ValueError: If x or y are not valid vertices.
        '''

        self._raise_error_if_invalid_vertex(x)
        self._raise_error_if_invalid_vertex(y)

        self.matrix[x][y] = self.no_edge_value

        if not self.directed:
            self.matrix[y][x] = self.no_edge_value

    def add_vertex(self) -> None:
        '''
        Add an isolated vertex to the graph.
        '''

        self.num_vertices += 1

        for row in self.matrix:
            row.append(self.no_edge_value)

        self.matrix.append([self.no_edge_value for _ in range(self.num_vertices)])

    def remove_vertex(self, x: int) -> None:
        '''
        Remove a vertex from the graph.

        Args:
            x (int): The vertex to be removed.

        Raises:
            ValueError: If the vertex does not exist.
        '''

        self._raise_error_if_invalid_vertex(x)

        self.num_vertices -= 1

        for row in self.matrix:
            del row[x]

        del self.matrix[x]

    def get_edge_weight(self, x: int, y: int) -> float:
        '''
        Get the weight of an edge.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.

        Returns:
            out (float): The weight of the edge.

        Raises:
            ValueError: If x or y are not valid vertices.
        '''

        self._raise_error_if_invalid_vertex(x)
        self._raise_error_if_invalid_vertex(y)

        return self.matrix[x][y]



'''
Binary Tree module.

This module contains the implementation of a binary tree.
'''

from typing import Any

class BinaryTree():
    '''
    Binary tree class.

    This class represents a binary tree.
    '''
    def __init__(self, data: Any):
        '''
        Initializes the tree.

        Args:
            data (Any): The data of the root node.
        '''
        self.data = data
        self.__left = None
        self.__right = None

    @property
    def left(self) -> 'BinaryTree':
        '''
        Returns the left child of the node.

        Returns:
            BinaryTree: The left child of the node.
        '''
        return self.__left

    @left.setter
    def left(self, left: 'BinaryTree') -> None:
        '''
        Sets the left child of the node.

        Args:
            left (BinaryTree): The left child of the node.
        '''
        self.__left = left

    @left.deleter
    def left(self) -> None:
        '''
        Deletes the left child of the node.
        '''
        self.__left = None

    @property
    def right(self) -> 'BinaryTree':
        '''
        Returns the right child of the node.

        Returns:
            BinaryTree: The right child of the node.
        '''
        return self.__right

    @right.setter
    def right(self, right: 'BinaryTree') -> None:
        '''
        Sets the right child of the node.

        Args:
            right (BinaryTree): The right child of the node.
        '''
        self.__right = right

    @right.deleter
    def right(self) -> None:
        '''
        Deletes the right child of the node.
        '''
        self.__right = None

    def is_leaf(self) -> bool:
        '''
        Checks whether the node is a leaf.

        Returns:
            bool: True if the node is a leaf, False otherwise.
        '''
        return self.__left is None and self.__right is None



import pytest
from pystrukts.tree import Trie, TrieNode

def test_trie_node_initialization():
    node = TrieNode()

    assert node.children == {}
    assert node.is_end_of_word == False

def test_trie_insert():
    trie = Trie()
    trie.insert("hello")

    assert trie.root.children['h'].children['e'].children['l'].children['l'].children['o'].is_end_of_word == True

def test_trie_search():
    trie = Trie()
    trie.insert("hello")

    assert trie.search("hello") == True
    assert trie.search("world") == False
    assert trie.search("hell") == False

def test_trie_delete():
    trie = Trie()
    trie.insert("hello")
    trie.insert("helloworld")

    trie.delete("hello")

    assert trie.search("hello") == False
    assert trie.search("helloworld") == True

    trie.delete("helloworld")
    assert trie.search("world") == False
    assert trie.root.children == {}

def test_trie_delete_empty():
    trie = Trie()
    trie.delete("hello")

    assert trie.root.children == {}

def test_trie_startswith():
    trie = Trie()
    trie.insert("hello")

    assert trie.startswith("hell") == True
    assert trie.startswith("world") == False
    assert trie.startswith("hello") == True


'''
Singly Linked List Module.

This module implements a Singly Linked List and Node data structures.
'''

from dataclasses import dataclass
from typing import Any
from .linked_list import LinkedList

@dataclass
class SinglyLinkedNode:
    '''
    SinglyLinkedNode Class.

    Attributes:
        data (Any): The data stored in the node.
        next (SinglyLinkedNode): The next node in the list.

    Methods:
        `__str__()`: Return the string representation of the node.
    '''

    data: Any = None
    next: 'SinglyLinkedNode' = None

    def __str__(self):
        return f"SinglyLinkedNode({self.data}){' -> ' + str(self.next) if self.next else ''}"

class SinglyLinkedList(LinkedList[SinglyLinkedNode]):
    '''
    SinglyLinkedList Class.

    Attributes:
        __head (SinglyLinkedNode): The head of the list.
        __tail (SinglyLinkedNode): The tail of the list.
        __length (int): The length of the list.
        __circular (bool): Whether the list is circular.

    Methods:
        `__str__()`: Return the string representation of the list.
        `__len__()`: Return the length of the list.
        `__iter__()`: Return an iterator for the list.
        `iscircular()`: Check if the list is circular.
        `_link(node1: Type[T], node2: Type[T])`: Link two nodes.
        `_preprocess_index(index: int)`: Preprocess the index.
        `insert(data: Any, index: int)`: Insert data at the specified index.
        `append(data: Any)`: Append data to the end of the list.
        `appendleft(data: Any)`: Append data to the beginning of the list.
        `remove(index: int)`: Remove data at the specified index.
        `pop()`: Remove data from the end of the list.
        `popleft()`: Remove data from the beginning of the list.
        `get(index: int)`: Get data at the specified index.
        `index(data: Any)`: Get the index of the specified data.
        `clear()`: Clear the list.
    '''

    def __init__(self, circular: bool = False):
        super().__init__(SinglyLinkedNode, circular)

    def _link(self, node1: SinglyLinkedNode, node2: SinglyLinkedNode):
        '''
        Link two nodes.
        Link two nodes.

        Args:
            node1 (SinglyLinkedNode): The first node.
            node2 (SinglyLinkedNode): The second node.
        '''

        node1.next = node2



import pytest
from pystrukts.graph import AdjacencyMatrix

def test_initilization():
    matrix = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    graph = AdjacencyMatrix(matrix=matrix)
    assert graph.num_vertices == 3
    assert graph.directed == True
    assert graph.matrix == [[0., 1., 2.], [3., 4., 5.], [6., 7., 8.]]

    graph2 = AdjacencyMatrix(matrix=matrix, directed=False)
    assert graph2.num_vertices == 3
    assert graph2.directed == False
    assert graph2.matrix == [[0., 1., 2.], [1., 4., 5.], [2., 5., 8.]] # _copy_upper

    graph3 = AdjacencyMatrix(3)
    assert graph3.num_vertices == 3
    assert graph3.directed == True
    assert graph3.matrix == [[float('inf') for _ in range(3)] for _ in range(3)]

    graph4 = AdjacencyMatrix(3, no_edge_value=0)
    assert graph4.num_vertices == 3
    assert graph4.directed == True
    assert graph4.matrix == [[0 for _ in range(3)] for _ in range(3)]

    # num_vertices ignored
    with pytest.warns(UserWarning):
        graph5 = AdjacencyMatrix(2, matrix = matrix)
        assert graph5.num_vertices == 3
        assert graph5.directed == True
        assert graph5.matrix == matrix

def test_initialization_error():
    with pytest.raises(ValueError):
        graph = AdjacencyMatrix(-1)

    with pytest.raises(ValueError):
        matrix = [[0, 1, 2], [3], [6, 7, 8]]
        graph = AdjacencyMatrix(matrix = matrix)

    with pytest.raises(ValueError):
        graph = AdjacencyMatrix()

def test_copy():
    matrix = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)
    graph2 = graph.copy()
    assert graph2.num_vertices == 3
    assert graph2.directed == True
    assert graph2.matrix == matrix
    assert graph2.no_edge_value == 0

def test_has_edge():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]

    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)
    assert graph.has_edge(0, 0) == False
    assert graph.has_edge(0, 1) == True
    assert graph.has_edge(0, 2) == False
    assert graph.has_edge(1, 0) == False
    assert graph.has_edge(1, 1) == False
    assert graph.has_edge(1, 2) == False
    assert graph.has_edge(2, 0) == True
    assert graph.has_edge(2, 1) == False
    assert graph.has_edge(2, 2) == True

def test_has_edge_error():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]
    graph = AdjacencyMatrix(matrix=matrix)

    with pytest.raises(ValueError):
        graph.has_edge(0, 3)

    with pytest.raises(ValueError):
        graph.has_edge(-1, 0)

def test_neighbors():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]

    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)
    assert graph.neighbors(0) == [1]
    assert graph.neighbors(1) == []
    assert graph.neighbors(2) == [0, 2]

def test_neighbors_error():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]
    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)

    with pytest.raises(ValueError):
        graph.neighbors(3)

    with pytest.raises(ValueError):
        graph.neighbors(-1)

def test_add_edge():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]

    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)

    assert graph.has_edge(0, 2) == False
    graph.add_edge(0, 2)
    assert graph.has_edge(0, 2) == True
    assert graph.matrix == [[0., 1., 1.], [0., 0., 0.], [1., 0., 1.]]

    assert graph.has_edge(1, 2) == False
    graph.add_edge(1, 2, 25)
    assert graph.has_edge(1, 2) == True
    assert graph.matrix == [[0., 1., 1.], [0., 0., 25.], [1., 0., 1.]]

def test_add_edge_undirected():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]

    graph = AdjacencyMatrix(3, directed=False, no_edge_value=0)

    assert graph.has_edge(0, 2) == False
    assert graph.has_edge(2, 0) == False
    graph.add_edge(0, 2)
    assert graph.has_edge(0, 2) == True
    assert graph.has_edge(2, 0) == True
    assert graph.get_edge_weight(0, 2) == 1
    assert graph.get_edge_weight(2, 0) == 1

    assert graph.has_edge(1, 2) == False
    assert graph.has_edge(2, 1) == False
    graph.add_edge(1, 2, 25)
    assert graph.has_edge(1, 2) == True
    assert graph.has_edge(2, 1) == True
    assert graph.get_edge_weight(1, 2) == 25
    assert graph.get_edge_weight(2, 1) == 25

def test_add_edge_error():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]
    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)

    with pytest.raises(ValueError):
        graph.add_edge(0, 3)

    with pytest.raises(ValueError):
        graph.add_edge(-1, 0)

def test_remove_edge():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]

    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)

    assert graph.has_edge(0, 1) == True
    graph.remove_edge(0, 1)
    assert graph.has_edge(0, 1) == False
    assert graph.matrix == [[0., 0., 0.], [0., 0., 0.], [1., 0., 1.]]

    assert graph.has_edge(2, 0) == True
    graph.remove_edge(2, 0)
    assert graph.has_edge(2, 0) == False
    assert graph.matrix == [[0., 0., 0.], [0., 0., 0.], [0., 0., 1.]]

def test_remove_edge_undirected():
    matrix = [[1, 1, 0], [1, 1, 0], [0, 0, 0]]
    graph = AdjacencyMatrix(matrix=matrix, directed=False, no_edge_value=0)

    assert graph.has_edge(0, 1) == True
    assert graph.has_edge(1, 0) == True
    graph.remove_edge(0, 1)
    assert graph.has_edge(0, 1) == False
    assert graph.has_edge(1, 0) == False
    assert graph.matrix == [[1., 0., 0.], [0., 1., 0.], [0., 0., 0.]]

    graph.remove_edge(0, 0)
    assert graph.matrix == [[0., 0., 0.], [0., 1., 0.], [0., 0., 0.]]

def test_remove_edge_error():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]
    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)

    with pytest.raises(ValueError):
        graph.remove_edge(0, 3)

    with pytest.raises(ValueError):
        graph.remove_edge(-1, 0)

def test_add_vertex():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]

    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)
    assert graph.num_vertices == 3
    assert graph.matrix == [[0., 1., 0.], [0., 0., 0.], [1., 0., 1.]]

    graph.add_vertex()
    assert graph.num_vertices == 4
    assert graph.matrix == [[0., 1., 0., 0.], [0., 0., 0., 0.], [1., 0., 1., 0.], [0., 0., 0., 0.]]

def test_remove_vertex():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]

    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)
    assert graph.num_vertices == 3
    assert graph.matrix == [[0., 1., 0.], [0., 0., 0.], [1., 0., 1.]]

    graph.remove_vertex(1)
    assert graph.num_vertices == 2
    assert graph.matrix == [[0., 0.], [1., 1.]]

    graph.remove_vertex(1)
    assert graph.num_vertices == 1
    assert graph.matrix == [[0.]]

def test_remove_vertex_error():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]
    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)

    with pytest.raises(ValueError):
        graph.remove_vertex(3)

    with pytest.raises(ValueError):
        graph.remove_vertex(-1)

def test_get_edge_weight():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]

    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)
    assert graph.get_edge_weight(0, 1) == 1
    assert graph.get_edge_weight(1, 0) == 0
    assert graph.get_edge_weight(2, 0) == 1

def test_get_edge_weight_error():
    matrix = [[0, 1, 0], [0, 0, 0], [1, 0, 1]]
    graph = AdjacencyMatrix(matrix=matrix, no_edge_value=0)

    with pytest.raises(ValueError):
        graph.get_edge_weight(0, 3)

    with pytest.raises(ValueError):
        graph.get_edge_weight(-1, 0)



import pytest
from pystrukts import MaxHeap

def test_heap_initialization():
    heap = MaxHeap([3, 1, 5, 7, 2])
    assert heap.peek() == 7

def test_empty_heap_initialization():
    heap = MaxHeap()
    assert not heap.data

def test_heap_push():
    heap = MaxHeap([3, 1, 5])
    heap.push(10)
    assert heap.peek() == 10

def test_push_negative_values():
    heap = MaxHeap([-1, 2, -3])
    heap.push(-10)
    assert heap.peek() == 2

def test_heap_pop():
    heap = MaxHeap([10, 20, 15, 2, 5])
    max_value = heap.pop()
    assert max_value == 20
    assert heap.peek() == 15

def test_pop_empty_heap():
    heap = MaxHeap()
    with pytest.raises(IndexError):
        heap.pop()

def test_heap_peek():
    heap = MaxHeap([1, 3, 2])
    assert heap.peek() == 3

def test_peek_empty_heap():
    heap = MaxHeap()
    with pytest.raises(IndexError):
        heap.peek()

def test_heap_str():
    heap = MaxHeap([4, 2, 9])
    assert str(heap) == "MaxHeap([9, 2, 4])"

def test_str_empty_heap():
    heap = MaxHeap()
    assert str(heap) == "MaxHeap([])"



'''
Trie Module.

This module implements a Trie (Prefix Tree) and TrieNode data structures.
'''

from collections import defaultdict
from dataclasses import dataclass

@dataclass
class TrieNode:
    '''
    TrieNode Class.

    Attributes:
        children (defaultdict): The children nodes of the node.
        is_end_of_word (bool): Whether the node is the end of a word.

    Methods:
        `__str__()`: Return the string representation of the node.
    '''

    children: defaultdict
    is_end_of_word: bool

    def __init__(self):
        self.children = defaultdict(TrieNode)
        self.is_end_of_word = False

@dataclass
class Trie:
    '''
    Trie Class.

    Attributes:
        root (TrieNode): The root node of the trie.

    Methods:
        `insert(word: str)`: Insert a word into the trie.
        `search(word: str)`: Search for a word in the trie.
        `delete(word: str)`: Delete a word from the trie.
        `startswith(prefix: str)`: Check if the trie contains a prefix.
    '''
    root = TrieNode()

    def insert(self, word: str) -> None:
        '''
        Insert a word into the trie.

        Args:
            word (str): The word to insert.
        '''

        node = self.root
        for char in word:
            node = node.children[char]
        node.is_end_of_word = True

    def search(self, word: str) -> bool:
        '''
        Search for a word in the trie.

        Args:
            word (str): The word to search.

        Returns:
            out (bool): Whether the word is in the trie.
        '''

        node = self.root
        for char in word:
            if char not in node.children:
                return False
            node = node.children[char]
        return node.is_end_of_word

    def _delete(self, node: TrieNode, word: str, depth: int) -> bool:
        '''
        Delete a node from the trie. Helper function for `delete`.

        Args:
            node (TrieNode): The current node.
            word (str): The word to delete.
            depth (int): The current depth.

        Returns:
            out (bool): Whether the node should be deleted.
        '''
        if depth == len(word):
            node.is_end_of_word = False
            return len(node.children) == 0

        char = word[depth]
        if char not in node.children:
            return False

        if self._delete(node.children[char], word, depth + 1):
            del node.children[char]
            return len(node.children) == 0

        return False

    def delete(self, word: str) -> None:
        '''
        Delete a word from the trie.

        Args:
            word (str): The word to delete.
        '''
        self._delete(self.root, word, 0)

    def startswith(self, prefix: str) -> bool:
        '''
        Check if the trie contains a prefix.

        Args:
            prefix (str): The prefix to check.

        Returns:
            out (bool): Whether the trie contains the prefix.
        '''
        node = self.root
        for char in prefix:
            if char not in node.children:
                return False
            node = node.children[char]
        return True



'''
Module for the Graph abstract class.
'''

from abc import ABCMeta, abstractmethod

class Graph(metaclass=ABCMeta):
    '''
    Abstract class for graph representations.

    Methods:
        copy() -> Graph: Create a copy of the graph.
        has_edge(x: int, y: int) -> bool: Check if there is an edge between two vertices.
        neighbors(x: int) -> list: Get the neighbors of a vertex.
        add_edge(x: int, y: int, w: float = 1.0) -> None: Add an edge between two vertices. 
            If the edge already exists, the weight is updated.
        remove_edge(x: int, y: int) -> None: Remove an edge between two vertices.
        add_vertex() -> None: Add an isolated vertex to the graph.
        remove_vertex(x: int) -> None: Remove a vertex from the graph.
    '''

    @abstractmethod
    def copy(self):
        '''
        Create a copy of the graph.

        Returns:
            Graph: A copy of the graph.
        '''

    @abstractmethod
    def has_edge(self, x: int, y: int) -> bool:
        '''
        Check if there is an edge between two vertices.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.

        Returns:
            out (bool): True if there is an edge between the vertices, False otherwise.
        '''

    @abstractmethod
    def neighbors(self, x: int) -> list:
        '''
        Get the neighbors of a vertex.

        Args:
            x (int): The vertex.

        Returns:
            out (list): The neighbors of the vertex.
        '''

    @abstractmethod
    def add_edge(self, x: int, y: int, w: float = 1.0):
        '''
        Add an edge between two vertices.
        
        Args:
            x (int): The source vertex.
            y (int): The destination vertex.
            w (float): The weight of the edge.
        '''

    @abstractmethod
    def remove_edge(self, x: int, y: int):
        '''
        Remove an edge between two vertices.

        Args:
            x (int): The source vertex.
            y (int): The destination vertex.
        '''

    @abstractmethod
    def add_vertex(self):
        '''
        Add an isolated vertex to the graph.
        '''

    @abstractmethod
    def remove_vertex(self, x: int):
        '''
        Remove a vertex from the graph.

        Args:
            x (int): The vertex.
        '''



import pytest
from pystrukts.tree import FenwickTree, RUPQ, RURQ


def test_initialization_ftree():
    f = [0, 1, 0, 1, 2, 3, 2, 1, 1, 0]
    ft = FenwickTree(f)
    assert ft.n == 10
    assert ft.ftree == [0, 0, 1, 0, 2, 2, 5, 2, 10, 1, 1]
    assert ft.query(1, 6) == 7
    assert ft.query(1, 3) == 1
    assert ft.select(7) == 6
    ft.update(5, 1)
    assert ft.query(1, 10) == 12

def test_rupq():
    r = RUPQ(10)
    r.update(2, 9, 7)
    r.update(6, 7, 3)
    assert r.query(1) == 0
    assert r.query(2) == 7
    assert r.query(3) == 7
    assert r.query(4) == 7
    assert r.query(5) == 7
    assert r.query(6) == 10
    assert r.query(7) == 10
    assert r.query(8) == 7
    assert r.query(9) == 7
    assert r.query(10) == 0

def test_rurq():
    r = RURQ(10)
    r.update(2, 9, 7)
    r.update(6, 7, 3)
    assert r.query(3, 5) == 21
    assert r.query(7, 8) == 17


import pytest
from pystrukts.tree import BinarySearchTree

def _inorder(tree: 'BinarySearchTree'):
    if tree is None:
        return []

    return _inorder(tree.left) + [tree.data] + _inorder(tree.right)


def test_initialization():
    tree = BinarySearchTree(1)
    assert tree.data == 1
    assert tree.left is None
    assert tree.right is None

    tree2 = BinarySearchTree(2)
    assert tree2.data == 2
    assert tree2.left is None
    assert tree2.right is None

def test_initialization_error():
    with pytest.raises(TypeError):
        BinarySearchTree()

def test_add_child():
    tree = BinarySearchTree(1)
    tree.left = BinarySearchTree(2)
    tree.right = BinarySearchTree(3)

    assert tree.left.data == 2
    assert tree.right.data == 3
    

def test_remove_child():
    tree = BinarySearchTree(1)
    tree.left = BinarySearchTree(2)
    tree.right = BinarySearchTree(3)

    del tree.left
    assert tree.left is None

    tree.right = None
    assert tree.right is None

def test_is_leaf():
    tree = BinarySearchTree(1)
    assert tree.is_leaf() == True

    tree.left = BinarySearchTree(2)
    assert tree.is_leaf() == False

    tree.left.left = BinarySearchTree(3)
    assert tree.left.is_leaf() == False
    assert tree.is_leaf() == False

    del tree.left.left
    assert tree.left.is_leaf() == True
    assert tree.is_leaf() == False

    tree.left = None
    assert tree.is_leaf() == True

def test_insert():
    tree = BinarySearchTree(1)
    tree.insert(2)
    tree.insert(3)
    tree.insert(0)
    tree.insert(-1)
    tree.insert(5)

    assert _inorder(tree) == [-1, 0, 1, 2, 3, 5]

def test_search():
    tree = BinarySearchTree(1)
    tree.insert(2)
    tree.insert(3)
    tree.insert(0)
    tree.insert(-1)
    tree.insert(5)

    assert tree.search(1) == True
    assert tree.search(2) == True
    assert tree.search(3) == True
    assert tree.search(0) == True
    assert tree.search(-1) == True
    assert tree.search(5) == True
    assert tree.search(4) == False
    assert tree.search(-2) == False

    tree.delete(1)
    assert tree.search(1) == False
    assert tree.search(2) == True
    assert tree.search(3) == True
    assert tree.search(0) == True
    assert tree.search(-1) == True
    assert tree.search(5) == True

def test_delete():
    tree = BinarySearchTree(1)
    tree.insert(5)
    tree.insert(0)
    tree.insert(-1)
    tree.insert(4)
    tree.insert(3)

    tree = tree.delete(1)
    assert _inorder(tree) == [-1, 0, 3, 4, 5]
    assert tree.data == 3

    tree = tree.delete(0)
    assert _inorder(tree) == [-1, 3, 4, 5]

    tree = tree.delete(5)
    assert _inorder(tree) == [-1, 3, 4]

    tree = tree.delete(3)
    assert _inorder(tree) == [-1, 4]

    tree = tree.delete(4)
    assert _inorder(tree) == [-1]

    tree = tree.delete(-1)
    assert _inorder(tree) == []




import pytest
from pystrukts.list import SinglyLinkedList, SinglyLinkedNode

def test_node_initialization():
    node = SinglyLinkedNode(1)

    assert node.data == 1
    assert node.next is None

def test_node_initialization_empty():
    node = SinglyLinkedNode()

    assert node.data is None
    assert node.next is None

def test_node_str():
    node = SinglyLinkedNode(1)

    assert str(node) == "SinglyLinkedNode(1)"

def test_node_str_empty():
    node = SinglyLinkedNode()

    assert str(node) == "SinglyLinkedNode(None)"

def test_node_str_multiple():
    node1 = SinglyLinkedNode(1)
    node2 = SinglyLinkedNode(2, node1)

    assert str(node2) == "SinglyLinkedNode(2) -> SinglyLinkedNode(1)"

def test_list_initialization():
    linked_list = SinglyLinkedList()

    assert len(linked_list) == 0
    assert linked_list.iscircular() == False
    assert linked_list._LinkedList__head is None
    assert linked_list._LinkedList__tail is None

def test_list_str():
    linked_list = SinglyLinkedList()

    assert str(linked_list) == "SinglyLinkedList(None)"

def test_list_iter_for():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert [x for x in linked_list] == [1, 2, 3]

def test_list_iter_while():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    iterator = iter(linked_list)
    assert next(iterator) == 1
    assert next(iterator) == 2
    assert next(iterator) == 3

    with pytest.raises(StopIteration):
        next(iterator)

def test_list_append():
    linked_list = SinglyLinkedList()
    linked_list.append(1)

    assert len(linked_list) == 1
    assert linked_list.get(0) == 1

def test_list_append_multiple():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert len(linked_list) == 3
    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 2
    assert linked_list.get(2) == 3

def test_list_prepend():
    linked_list = SinglyLinkedList()
    linked_list.appendleft(1)

    assert len(linked_list) == 1
    assert linked_list.get(0) == 1

def test_list_prepend_multiple():
    linked_list = SinglyLinkedList()
    linked_list.appendleft(1)
    linked_list.appendleft(2)
    linked_list.appendleft(3)

    assert len(linked_list) == 3
    assert linked_list.get(0) == 3
    assert linked_list.get(1) == 2
    assert linked_list.get(2) == 1

def test_insert():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(4)
    linked_list.insert(3, 2)

    assert len(linked_list) == 4
    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 2
    assert linked_list.get(2) == 3
    assert linked_list.get(3) == 4

def test_insert_start():
    linked_list = SinglyLinkedList()
    linked_list.append(2)
    linked_list.append(3)
    linked_list.insert(1, 0)

    assert len(linked_list) == 3
    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 2
    assert linked_list.get(2) == 3

def test_insert_end():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.insert(3, 2)

    assert len(linked_list) == 3
    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 2
    assert linked_list.get(2) == 3

def test_insert_negative_index():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.insert(3, -1)

    assert len(linked_list) == 3
    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 2
    assert linked_list.get(2) == 3

def test_insert_out_of_bounds():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)

    with pytest.raises(IndexError): 
        linked_list.insert(3, 5)

def test_pop():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert linked_list.pop() == 3
    assert len(linked_list) == 2
    assert linked_list.get(-1) == 2

def test_pop_last():
    linked_list = SinglyLinkedList()
    linked_list.append(1)

    assert linked_list.pop() == 1
    assert len(linked_list) == 0
    assert linked_list._LinkedList__head is None
    assert linked_list._LinkedList__tail is None

def test_pop_empty():
    linked_list = SinglyLinkedList()

    with pytest.raises(IndexError): 
        linked_list.pop()

def test_popleft():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert linked_list.popleft() == 1
    assert len(linked_list) == 2
    assert linked_list.get(0) == 2

def test_popleft_last():
    linked_list = SinglyLinkedList()
    linked_list.append(1)

    assert linked_list.popleft() == 1
    assert len(linked_list) == 0
    assert linked_list._LinkedList__head is None
    assert linked_list._LinkedList__tail is None

def test_popleft_empty():
    linked_list = SinglyLinkedList()

    with pytest.raises(IndexError): 
        linked_list.popleft()

def test_remove():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)
    linked_list.append(4)
    linked_list.remove(2)

    assert len(linked_list) == 3
    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 2
    assert linked_list.get(2) == 4

def test_remove_start():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)
    linked_list.remove(0)

    assert len(linked_list) == 2
    assert linked_list.get(0) == 2
    assert linked_list.get(1) == 3

def test_remove_end():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)
    linked_list.remove(2)

    assert len(linked_list) == 2
    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 2

def test_remove_negative_index():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)
    linked_list.remove(-2)

    assert len(linked_list) == 2
    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 3

def test_remove_out_of_bounds():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    with pytest.raises(IndexError): 
        linked_list.remove(5)

    with pytest.raises(IndexError): 
        linked_list.remove(-5)

def test_get():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert linked_list.get(0) == 1
    assert linked_list.get(1) == 2
    assert linked_list.get(2) == 3

def test_get_negative_index():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert linked_list.get(-1) == 3
    assert linked_list.get(-2) == 2
    assert linked_list.get(-3) == 1

def test_get_out_of_bounds():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    with pytest.raises(IndexError): 
        linked_list.get(5)
    
    with pytest.raises(IndexError): 
        linked_list.get(-5)

def test_index():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert linked_list.index(1) == 0
    assert linked_list.index(2) == 1
    assert linked_list.index(3) == 2

def test_index_not_found():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    with pytest.raises(ValueError): 
        linked_list.index(4)

def test_clear():
    linked_list = SinglyLinkedList()
    linked_list.append(1)
    linked_list.append(2)
    linked_list.clear()

    assert len(linked_list) == 0

def test_is_circular():
    linked_list = SinglyLinkedList(circular=True)

    assert linked_list.iscircular() == True

def test_append_circular():
    linked_list = SinglyLinkedList(circular=True)
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert linked_list._LinkedList__head.data == 1
    assert linked_list._LinkedList__tail.data == 3
    assert linked_list._LinkedList__tail.next == linked_list._LinkedList__head

def test_appendleft_circular():
    linked_list = SinglyLinkedList(circular=True)
    linked_list.appendleft(1)
    linked_list.appendleft(2)
    linked_list.appendleft(3)

    assert linked_list._LinkedList__head.data == 3
    assert linked_list._LinkedList__tail.data == 1
    assert linked_list._LinkedList__tail.next == linked_list._LinkedList__head

def test_pop_circular():
    linked_list = SinglyLinkedList(circular=True)
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert linked_list.pop() == 3
    assert linked_list._LinkedList__head.data == 1
    assert linked_list._LinkedList__tail.data == 2
    assert linked_list._LinkedList__tail.next == linked_list._LinkedList__head

def test_popleft_circular():
    linked_list = SinglyLinkedList(circular=True)
    linked_list.append(1)
    linked_list.append(2)
    linked_list.append(3)

    assert linked_list.popleft() == 1
    assert linked_list._LinkedList__head.data == 2
    assert linked_list._LinkedList__tail.data == 3
    assert linked_list._LinkedList__tail.next == linked_list._LinkedList__head


import pytest
from pystrukts import UnionFind

def test_ufds_initialization():
    ufds = UnionFind(5)
    assert ufds.parents == [0, 1, 2, 3, 4]
    assert ufds.ranks == [0, 0, 0, 0, 0]
    assert ufds.sizes == [1, 1, 1, 1, 1]
    assert ufds.numdisjoint == 5

def test_find():
    ufds = UnionFind(5)
    assert ufds.find(3) == 3
    ufds.union(1, 3)
    assert ufds.find(3) == 1

def test_union():
    ufds = UnionFind(5)
    ufds.union(1, 3)

    assert ufds.parents == [0, 1, 2, 1, 4]
    assert ufds.ranks == [0, 1, 0, 0, 0]
    assert ufds.sizes == [1, 2, 1, 1, 1]
    assert ufds.numdisjoint == 4

    ufds.union(2, 4)

    assert ufds.parents == [0, 1, 2, 1, 2]
    assert ufds.ranks == [0, 1, 1, 0, 0]
    assert ufds.sizes == [1, 2, 2, 1, 1]
    assert ufds.numdisjoint == 3

    ufds.union(1, 4)

    assert ufds.parents == [0, 1, 1, 1, 2]
    assert ufds.ranks == [0, 2, 1, 0, 0]
    assert ufds.sizes == [1, 4, 2, 1, 1]
    assert ufds.numdisjoint == 2

    ufds.union(0, 2)

    assert ufds.parents == [1, 1, 1, 1, 2]
    assert ufds.ranks == [0, 2, 1, 0, 0]
    assert ufds.sizes == [1, 5, 2, 1, 1]
    assert ufds.numdisjoint == 1

def test_union_sets():
    ufds = UnionFind(5)
    ufds.union(1, 3)
    ufds.union(1, 4)

    assert ufds.find(1) != ufds.find(0)
    assert ufds.find(1) != ufds.find(2)
    assert ufds.find(1) == ufds.find(3)
    assert ufds.find(1) == ufds.find(4)
    assert ufds.numdisjoint == 3

def test_union_same_set():
    ufds = UnionFind(5)
    ufds.union(1, 3)
    ufds.union(1, 4)

    numdisjoint = ufds.numdisjoint
    parents = ufds.parents.copy()
    ranks = ufds.ranks.copy()
    sizes = ufds.sizes.copy()

    ufds.union(1, 4)

    assert ufds.parents == parents
    assert ufds.ranks == ranks
    assert ufds.sizes == sizes
    assert ufds.numdisjoint == numdisjoint

def test_size():
    ufds = UnionFind(4)
    ufds.union(1, 3)

    assert ufds.size(0) == 1
    assert ufds.size(1) == 2
    assert ufds.size(2) == 1
    assert ufds.size(3) == 2

    ufds.union(1, 2)

    assert ufds.size(0) == 1
    assert ufds.size(1) == 3
    assert ufds.size(2) == 3
    assert ufds.size(3) == 3

    ufds.union(0, 2)
    
    assert ufds.size(0) == 4
    assert ufds.size(1) == 4
    assert ufds.size(2) == 4
    assert ufds.size(3) == 4